@@ -41,26 +41,27 @@ SEXP cpp_set_rm_attr(SEXP x, SEXP which) {
 [[cpp11::register]]
 SEXP cpp_set_add_attributes(SEXP x, SEXP attributes, bool add) {
   int NP = 0;
-  if (Rf_isNull(attributes) ||
+  SEXP attrs = Rf_protect(Rf_isPairList(attributes) ? Rf_coerceVector(attributes, VECSXP) : attributes);
+  ++NP;
+  int n = Rf_length(attrs);
+  bool attrs_are_a_list = Rf_isVectorList(attrs);
+  if (Rf_isNull(attrs) ||
       // is.null or empty list?
-      (TYPEOF(attributes) == VECSXP && Rf_length(attributes) == 0)){
+      (attrs_are_a_list && n == 0)){
     if (add){
       return x;
     } else {
       return cpp_set_rm_attributes(x);
     }
   }
-  if (!add){
-    cpp_set_rm_attributes(x);
-  }
-  SEXP names = Rf_protect(Rf_getAttrib(attributes, R_NamesSymbol)); ++NP;
-  if (!Rf_isVectorList(attributes) || Rf_isNull(names)){
+  SEXP names = Rf_protect(Rf_getAttrib(attrs, R_NamesSymbol)); ++NP;
+  if (!attrs_are_a_list || Rf_isNull(names)){
     Rf_unprotect(NP);
     Rf_error("attributes must be a named list");
   }
-  const SEXP *p_attributes = VECTOR_PTR_RO(attributes);
+  if (!add) cpp_set_rm_attributes(x);
+  const SEXP *p_attributes = VECTOR_PTR_RO(attrs);
   const SEXP *p_names = STRING_PTR_RO(names);
-  int n = Rf_length(attributes);
   for (int i = 0; i < n; ++i){
     SEXP attr_nm = Rf_protect(Rf_installChar(p_names[i])); ++NP;
     if (r_address(x) == r_address(p_attributes[i])){
@@ -73,3 +74,20 @@ SEXP cpp_set_add_attributes(SEXP x, SEXP attributes, bool add) {
   Rf_unprotect(NP);
   return x;
 }
+
+// Copy attributes from source to target
+void cpp_copy_attributes(SEXP source, SEXP target, bool deep_copy){
+  SEXP target_attrs = Rf_protect(deep_copy ? Rf_duplicate(ATTRIB(source)) : ATTRIB(source));
+  cpp_set_add_attributes(target, target_attrs, false);
+  Rf_unprotect(1);
+}
+
+// Copy names from source to target
+void cpp_copy_names(SEXP source, SEXP target, bool deep_copy){
+  SEXP source_nms = Rf_protect(Rf_getAttrib(source, R_NamesSymbol));
+  SEXP target_nms = Rf_protect(deep_copy ? Rf_duplicate(source_nms) : source_nms);
+  if (!Rf_isNull(source_nms)){
+    Rf_setAttrib(target, R_NamesSymbol, target_nms);
+  }
+  Rf_unprotect(2);
+}
