@@ -13,14 +13,14 @@ void check_order_value(unsigned int x, unsigned int rng, int n_prot){
 SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
   R_xlen_t size = Rf_xlength(x);
   R_xlen_t fill_size = Rf_xlength(fill);
-  int n_protections = 0;
+  int NP = 0;
   if (fill_size > 1){
     Rf_error("fill size must be NULL or length 1");
   }
   bool set_and_altrep = set && ALTREP(x);
   SEXP out;
-  SEXP xvec = Rf_protect(altrep_materialise(x));
-  ++n_protections;
+  SEXP xvec = Rf_protect(set_and_altrep ? altrep_materialise(x) : x);
+  ++NP;
   if (set_and_altrep){
     Rf_warning("Cannot lag an ALTREP by reference, a copy has been made.\n\tEnsure the result is assigned to an object if used in further calculations\n\te.g. `x <- lag_(x, set = TRUE)`");
   }
@@ -37,197 +37,116 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
       fill_value = Rf_asInteger(fill);
     }
     out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
-    ++n_protections;
+    ++NP;
     int *p_out = INTEGER(out);
-    if (set){
-      R_xlen_t tempi;
-      int tempv;
-      // If k = 0 then no lag occurs
-      if (std::abs(k) >= 1){
-        SEXP lag_temp = Rf_protect(Rf_allocVector(INTSXP, std::abs(k)));
-        ++n_protections;
-        int* __restrict__ p_lag = INTEGER(lag_temp);
-        // Positive lags
-        if (k >= 0){
-          for (R_xlen_t i = 0; i < k; ++i) {
-            p_lag[i] = p_out[i];
-            p_out[i] = fill_value;
-          }
-          for (R_xlen_t i = k; i < size; ++i) {
-            tempi = ((i - k) % k);
-            tempv = p_lag[tempi];
-            p_lag[tempi] = p_out[i];
-            p_out[i] = tempv;
-          }
-          // Negative lags
-        } else {
-          for (R_xlen_t i = size - 1; i >= size + k; --i) {
-            p_lag[size - i - 1] = p_out[i];
-            p_out[i] = fill_value;
-          }
-          for (R_xlen_t i = size + k - 1; i >= 0; --i) {
-            tempi = ( (size - (i - k) - 1) % k);
-            tempv = p_lag[tempi];
-            p_lag[tempi] = p_out[i];
-            p_out[i] = tempv;
-          }
-        }
-
-      }
+    int *p_x = INTEGER(xvec);
+    if (k >= 0){
+      memmove(&p_out[k], &p_x[0], (size - k) * sizeof(int));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = 0; i < k; ++i) p_out[i] = fill_value;
     } else {
-      int *p_x = INTEGER(xvec);
-      if (k >= 0){
-        OMP_FOR_SIMD
-        for (R_xlen_t i = 0; i < k; ++i) p_out[i] = fill_value;
-        OMP_FOR_SIMD
-        for (R_xlen_t i = k; i < size; ++i) p_out[i] = p_x[i - k];
-      } else {
-        OMP_FOR_SIMD
-        for (R_xlen_t i = size - 1; i >= size + k; --i) p_out[i] = fill_value;
-        OMP_FOR_SIMD
-        for (R_xlen_t i = size + k - 1; i >= 0; --i) p_out[i] = p_x[i - k];
-      }
+      memmove(&p_out[0], &p_x[-k], (size + k) * sizeof(int));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = size - 1; i >= size + k; --i) p_out[i] = fill_value;
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
     break;
   }
   case REALSXP: {
+    if (is_int64(x)){
+    k = k >= 0 ? std::min(size, k) : std::max(-size, k);
+    long long int fill_value = NA_INTEGER64;
+    if (fill_size >= 1){
+      SEXP temp_fill = Rf_protect(
+        cpp_numeric_to_int64(
+          is_int64(fill) ? fill : Rf_coerceVector(fill, REALSXP)
+        )
+      ); ++NP;
+      fill_value = INTEGER64_PTR(temp_fill)[0];
+    }
+    out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
+    ++NP;
+    long long int *p_out = INTEGER64_PTR(out);
+    long long int *p_x = INTEGER64_PTR(xvec);
+
+    // sizeof(long long int) == sizeof(double), both are 8 bytes
+
+    if (k >= 0){
+      memmove(&p_out[k], &p_x[0], (size - k) * sizeof(long long int));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = 0; i < k; ++i) p_out[i] = fill_value;
+    } else {
+      memmove(&p_out[0], &p_x[-k], (size + k) * sizeof(long long int));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = size - 1; i >= size + k; --i) p_out[i] = fill_value;
+    }
+    if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
+      SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
+      ++NP;
+      SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
+      ++NP;
+      Rf_setAttrib(out, R_NamesSymbol, new_names);
+    }
+  } else {
     k = k >= 0 ? std::min(size, k) : std::max(-size, k);
     double fill_value = NA_REAL;
     if (fill_size >= 1){
       fill_value = Rf_asReal(fill);
     }
     out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
-    ++n_protections;
+    ++NP;
     double *p_out = REAL(out);
-    if (set){
-      R_xlen_t tempi;
-      double tempv;
-      // If k = 0 then no lag occurs
-      if (std::abs(k) >= 1){
-        SEXP lag_temp = Rf_protect(Rf_allocVector(REALSXP, std::abs(k)));
-        ++n_protections;
-        double* __restrict__ p_lag = REAL(lag_temp);
-        // Positive lags
-        if (k >= 0){
-          for (R_xlen_t i = 0; i < k; ++i) {
-            p_lag[i] = p_out[i];
-            p_out[i] = fill_value;
-          }
-          for (R_xlen_t i = k; i < size; ++i) {
-            tempi = ((i - k) % k);
-            tempv = p_lag[tempi];
-            p_lag[tempi] = p_out[i];
-            p_out[i] = tempv;
-          }
-          // Negative lags
-        } else {
-          for (R_xlen_t i = size - 1; i >= size + k; --i) {
-            p_lag[size - i - 1] = p_out[i];
-            p_out[i] = fill_value;
-          }
-          for (R_xlen_t i = size + k - 1; i >= 0; --i) {
-            tempi = ( (size - (i - k) - 1) % k);
-            tempv = p_lag[tempi];
-            p_lag[tempi] = p_out[i];
-            p_out[i] = tempv;
-          }
-        }
-
-      }
+    double *p_x = REAL(xvec);
+    if (k >= 0){
+      memmove(&p_out[k], &p_x[0], (size - k) * sizeof(double));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = 0; i < k; ++i) p_out[i] = fill_value;
     } else {
-      double *p_x = REAL(xvec);
-      if (k >= 0){
-        OMP_FOR_SIMD
-        for (R_xlen_t i = 0; i < k; ++i) p_out[i] = fill_value;
-        OMP_FOR_SIMD
-        for (R_xlen_t i = k; i < size; ++i) p_out[i] = p_x[i - k];
-      } else {
-        OMP_FOR_SIMD
-        for (R_xlen_t i = size - 1; i >= size + k; --i) p_out[i] = fill_value;
-        OMP_FOR_SIMD
-        for (R_xlen_t i = size + k - 1; i >= 0; --i) p_out[i] = p_x[i - k];
-      }
+      memmove(&p_out[0], &p_x[-k], (size + k) * sizeof(double));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = size - 1; i >= size + k; --i) p_out[i] = fill_value;
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
+  }
     break;
   }
   case CPLXSXP: {
     k = k >= 0 ? std::min(size, k) : std::max(-size, k);
     SEXP fill_sexp = Rf_protect(Rf_allocVector(CPLXSXP, 1));
-    ++n_protections;
+    ++NP;
     Rcomplex *p_fill = COMPLEX(fill_sexp);
     p_fill[0].i = NA_REAL;
     p_fill[0].r = NA_REAL;
     Rcomplex fill_value = fill_size >= 1 ? Rf_asComplex(fill) : COMPLEX(fill_sexp)[0];
     out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
-    ++n_protections;
+    ++NP;
     Rcomplex *p_out = COMPLEX(out);
-    if (set){
-      R_xlen_t tempi;
-      Rcomplex tempv;
-      // If k = 0 then no lag occurs
-      if (std::abs(k) >= 1){
-        SEXP lag_temp = Rf_protect(Rf_allocVector(CPLXSXP, std::abs(k)));
-        ++n_protections;
-        Rcomplex* __restrict__ p_lag = COMPLEX(lag_temp);
-        // Positive lags
-        if (k >= 0){
-          for (R_xlen_t i = 0; i < k; ++i) {
-            p_lag[i].i = p_out[i].i;
-            p_lag[i].r = p_out[i].r;
-            SET_COMPLEX_ELT(out, i, fill_value);
-          }
-          for (R_xlen_t i = k; i < size; ++i) {
-            tempi = ((i - k) % k);
-            tempv = p_lag[tempi];
-            SET_COMPLEX_ELT(lag_temp, tempi, p_out[i]);
-            SET_COMPLEX_ELT(out, i, tempv);
-          }
-          // Negative lags
-        } else {
-          for (R_xlen_t i = size - 1; i >= size + k; --i) {
-            SET_COMPLEX_ELT(lag_temp, size - i - 1, p_out[i]);
-            SET_COMPLEX_ELT(out, i, fill_value);
-          }
-          for (R_xlen_t i = size + k - 1; i >= 0; --i) {
-            tempi = ( (size - (i - k) - 1) % k);
-            tempv = p_lag[tempi];
-            SET_COMPLEX_ELT(lag_temp, tempi, p_out[i]);
-            SET_COMPLEX_ELT(out, i, tempv);
-          }
-        }
-
-      }
+    Rcomplex *p_x = COMPLEX(xvec);
+    if (k >= 0){
+      memmove(&p_out[k], &p_x[0], (size - k) * sizeof(Rcomplex));
+      for (R_xlen_t i = 0; i < k; ++i) SET_COMPLEX_ELT(out, i, fill_value);
     } else {
-      Rcomplex *p_x = COMPLEX(xvec);
-      if (k >= 0){
-        for (R_xlen_t i = 0; i < size; ++i) {
-          SET_COMPLEX_ELT(out, i, i >= k ? p_x[i - k] : fill_value);
-        }
-      } else {
-        for (R_xlen_t i = size - 1; i >= 0; --i) {
-          SET_COMPLEX_ELT(out, i, (i - size) < k ? p_x[i - k] : fill_value);
-        }
-      }
+      memmove(&p_out[0], &p_x[-k], (size + k) * sizeof(Rcomplex));
+      OMP_FOR_SIMD
+      for (R_xlen_t i = size - 1; i >= size + k; --i) SET_COMPLEX_ELT(out, i, fill_value);
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
     break;
@@ -235,18 +154,18 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
   case STRSXP: {
     k = k >= 0 ? std::min(size, k) : std::max(-size, k);
     SEXP fill_char = Rf_protect(fill_size >= 1 ? Rf_asChar(fill) : NA_STRING);
-    ++n_protections;
+    ++NP;
     out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
-    ++n_protections;
+    ++NP;
     const SEXP *p_out = STRING_PTR_RO(out);
     if (set){
       R_xlen_t tempi;
       // If k = 0 then no lag occurs
       if (std::abs(k) >= 1){
         SEXP lag_temp = Rf_protect(Rf_allocVector(STRSXP, std::abs(k)));
-        ++n_protections;
+        ++NP;
         SEXP tempv = Rf_protect(Rf_allocVector(STRSXP, 1));
-        ++n_protections;
+        ++NP;
         const SEXP* __restrict__ p_lag = STRING_PTR_RO(lag_temp);
         // Positive lags
         if (k >= 0){
@@ -286,9 +205,9 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
     break;
@@ -297,16 +216,16 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
     k = k >= 0 ? std::min(size, k) : std::max(-size, k);
     SEXP raw_sexp = Rf_protect(Rf_coerceVector(fill, RAWSXP));
     Rbyte fill_raw = fill_size == 0 ? RAW(Rf_ScalarRaw(0))[0] : RAW(raw_sexp)[0];
-    ++n_protections;
+    ++NP;
     out = Rf_protect(set ? xvec : Rf_duplicate(xvec));
-    ++n_protections;
+    ++NP;
     Rbyte *p_out = RAW(out);
     if (set){
       R_xlen_t tempi;
       // If k = 0 then no lag occurs
       if (std::abs(k) >= 1){
         SEXP lag_temp = Rf_protect(Rf_allocVector(RAWSXP, std::abs(k)));
-        ++n_protections;
+        ++NP;
         Rbyte *p_lag = RAW(lag_temp);
         // Positive lags
         if (k >= 0){
@@ -348,9 +267,9 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
     break;
@@ -359,7 +278,7 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
     if (recursive){
     const SEXP *p_x = VECTOR_PTR_RO(xvec);
     out = Rf_protect(Rf_allocVector(VECSXP, size));
-    ++n_protections;
+    ++NP;
     SHALLOW_DUPLICATE_ATTRIB(out, xvec);
     for (R_xlen_t i = 0; i < size; ++i){
       SET_VECTOR_ELT(out, i, cpp_lag(p_x[i], k, fill, set, true));
@@ -367,18 +286,18 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
   } else {
     k = k >= 0 ? std::min(size, k) : std::max(-size, k);
     SEXP fill_value = Rf_protect(Rf_coerceVector(fill_size >= 1 ? fill : R_NilValue, VECSXP));
-    ++n_protections;
+    ++NP;
     out = Rf_protect(set ? xvec : Rf_allocVector(VECSXP, size));
-    ++n_protections;
+    ++NP;
     const SEXP *p_out = VECTOR_PTR_RO(out);
     if (set){
       R_xlen_t tempi;
       // If k = 0 then no lag occurs
       if (std::abs(k) >= 1){
         SEXP lag_temp = Rf_protect(Rf_allocVector(VECSXP, std::abs(k)));
-        ++n_protections;
+        ++NP;
         SEXP tempv = Rf_protect(Rf_allocVector(VECSXP, 1));
-        ++n_protections;
+        ++NP;
         SEXP* __restrict__ p_lag = VECTOR_PTR(lag_temp);
         // Positive lags
         if (k >= 0){
@@ -420,20 +339,20 @@ SEXP cpp_lag(SEXP x, R_xlen_t k, SEXP fill, bool set, bool recursive) {
     }
     if (!Rf_isNull(Rf_getAttrib(xvec, R_NamesSymbol))){
       SEXP old_names = Rf_protect(Rf_getAttrib(xvec, R_NamesSymbol));
-      ++n_protections;
+      ++NP;
       SEXP new_names = Rf_protect(cpp_lag(old_names, k, R_NilValue, set, recursive));
-      ++n_protections;
+      ++NP;
       Rf_setAttrib(out, R_NamesSymbol, new_names);
     }
   }
   break;
   }
   default: {
-    Rf_unprotect(n_protections);
+    Rf_unprotect(NP);
     Rf_error("%s cannot handle an object of type %s", __func__, Rf_type2char(TYPEOF(xvec)));
   }
   }
-  Rf_unprotect(n_protections);
+  Rf_unprotect(NP);
   return out;
 }
 
@@ -443,7 +362,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
   int rl_size = Rf_length(run_lengths);
   int lag_size = Rf_length(lag);
   int fill_size = Rf_length(fill);
-  int n_protections = 0;
+  int NP = 0;
   if (fill_size > 1){
     Rf_error("fill size must be NULL or length 1");
   }
@@ -464,15 +383,15 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
   // if they are NULL
 
   SEXP dummy_vec1 = Rf_protect(Rf_allocVector(INTSXP, 0));
-  ++n_protections;
+  ++NP;
   SEXP dummy_vec2 = Rf_protect(Rf_allocVector(INTSXP, 0));
-  ++n_protections;
+  ++NP;
   Rf_protect(lag = Rf_coerceVector(lag, INTSXP));
-  ++n_protections;
+  ++NP;
   Rf_protect(order = has_order ? Rf_coerceVector(order, INTSXP) : R_NilValue);
-  ++n_protections;
+  ++NP;
   Rf_protect(run_lengths = has_rl ? Rf_coerceVector(run_lengths, INTSXP) : R_NilValue);
-  ++n_protections;
+  ++NP;
   // std::variant<int*, double*> p_o;
   // typedef std::conditional<true, int*, double*>::type p_o;
   int *p_o = INTEGER(has_order ? order : dummy_vec1);
@@ -488,14 +407,14 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
   if (!has_rl) rl_size = 1;
 
   // Macro to lag names(x)
-#define LAG_R_NAMES                                            \
-  if (!Rf_isNull(Rf_getAttrib(x, R_NamesSymbol))){             \
-    SEXP old_names = Rf_protect(Rf_getAttrib(x, R_NamesSymbol));\
-    ++n_protections;                                           \
+#define CHEAPR_LAG_R_NAMES                                                                                  \
+  if (!Rf_isNull(Rf_getAttrib(x, R_NamesSymbol))){                                                   \
+    SEXP old_names = Rf_protect(Rf_getAttrib(x, R_NamesSymbol));                                     \
+    ++NP;                                                                                            \
     SEXP new_names = Rf_protect(cpp_lag2(old_names, lag, order, run_lengths, R_NilValue, recursive));\
-    ++n_protections;                                           \
-    Rf_setAttrib(out, R_NamesSymbol, new_names);               \
-  }                                                            \
+    ++NP;                                                                                            \
+    Rf_setAttrib(out, R_NamesSymbol, new_names);                                                     \
+  }                                                                                                  \
 
   // Initialise range of possible order values to fast check
   // user-supplied order values
@@ -518,7 +437,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       fill_value = Rf_asInteger(fill);
     }
     out = Rf_protect(Rf_duplicate(x));
-    ++n_protections;
+    ++NP;
     int *p_out = INTEGER(out);
     for (int i = 0; i != rl_size; ++i){
       run_start = run_end; // Start at the end of the previous run
@@ -528,13 +447,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
 
       // If any run-lengths are negative (or NA) we stop
       if (rl < 0){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("All run lengths must be non-NA and >= 0");
       }
 
       // If the cumulative run-length exceeds length(x) we stop
       if (run_end > size){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
       // int trick to calculate inclusive bounded between(x, lo, hi)
@@ -544,7 +463,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
         // Check that order value is valid
         if (has_order){
           oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
+          check_order_value(oi, o_rng, NP);
         } else {
           oi = j;
         }
@@ -561,10 +480,10 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       }
     }
     if (run_end != size){
-      Rf_unprotect(n_protections);
+      Rf_unprotect(NP);
       Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
     }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
     break;
   }
   case REALSXP: {
@@ -572,52 +491,105 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     if (has_order && (size != o_size)){
       Rf_error("length(order) must equal length(x) (%d)", size);
     }
-    double *p_x = REAL(x);
-    double fill_value = NA_REAL;
-    if (fill_size >= 1){
-      fill_value = Rf_asReal(fill);
-    }
-    out = Rf_protect(Rf_duplicate(x));
-    ++n_protections;
-    double *p_out = REAL(out);
-    for (int i = 0; i != rl_size; ++i){
-      run_start = run_end; // Start at the end of the previous run
-      rl = has_rl ? p_rl[i] : size; // Current run-length
-      run_end += rl; // Cumulative run-length
-
-      // If any run-lengths are negative (or NA) we stop
-      if (rl < 0){
-        Rf_unprotect(n_protections);
-        Rf_error("All run lengths must be non-NA and >= 0");
+    if (is_int64(x)){
+      long long int *p_x = INTEGER64_PTR(x);
+      long long int fill_value = NA_INTEGER64;
+      if (fill_size >= 1){
+        SEXP temp_fill = Rf_protect(
+          cpp_numeric_to_int64(
+            is_int64(fill) ? fill : Rf_coerceVector(fill, REALSXP)
+          )
+        ); ++NP;
+        fill_value = INTEGER64_PTR(temp_fill)[0];
       }
+      out = Rf_protect(Rf_duplicate(x));
+      ++NP;
+      long long int *p_out = INTEGER64_PTR(out);
+      for (int i = 0; i != rl_size; ++i){
+        run_start = run_end; // Start at the end of the previous run
+        rl = has_rl ? p_rl[i] : size; // Current run-length
+        run_end += rl; // Cumulative run-length
 
-      // If the cumulative run-length exceeds length(x) we stop
-      if (run_end > size){
-        Rf_unprotect(n_protections);
+        // If any run-lengths are negative (or NA) we stop
+        if (rl < 0){
+          Rf_unprotect(NP);
+          Rf_error("All run lengths must be non-NA and >= 0");
+        }
+
+        // If the cumulative run-length exceeds length(x) we stop
+        if (run_end > size){
+          Rf_unprotect(NP);
+          Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
+        }
+        // Loop starting from the end of the previous run-length
+        for (int j = run_start; j != run_end; ++j){
+          // Check that order value is valid
+          if (has_order){
+            oi = p_o[j] - 1;
+            check_order_value(oi, o_rng, NP);
+          } else {
+            oi = j;
+          }
+          k = recycle_lag ? p_lag[oi % lag_size] : lag1;
+          if (k >= 0){
+            p_out[oi] = (j - run_start) >= k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
+          } else {
+            p_out[oi] = (j - run_end) < k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
+          }
+        }
+      }
+      if (run_end != size){
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
-      // Loop starting from the end of the previous run-length
-      for (int j = run_start; j != run_end; ++j){
-        // Check that order value is valid
-        if (has_order){
-          oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
-        } else {
-          oi = j;
+    } else {
+      double *p_x = REAL(x);
+      double fill_value = NA_REAL;
+      if (fill_size >= 1){
+        fill_value = Rf_asReal(fill);
+      }
+      out = Rf_protect(Rf_duplicate(x));
+      ++NP;
+      double *p_out = REAL(out);
+      for (int i = 0; i != rl_size; ++i){
+        run_start = run_end; // Start at the end of the previous run
+        rl = has_rl ? p_rl[i] : size; // Current run-length
+        run_end += rl; // Cumulative run-length
+
+        // If any run-lengths are negative (or NA) we stop
+        if (rl < 0){
+          Rf_unprotect(NP);
+          Rf_error("All run lengths must be non-NA and >= 0");
         }
-        k = recycle_lag ? p_lag[oi % lag_size] : lag1;
-        if (k >= 0){
-          p_out[oi] = (j - run_start) >= k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
-        } else {
-          p_out[oi] = (j - run_end) < k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
+
+        // If the cumulative run-length exceeds length(x) we stop
+        if (run_end > size){
+          Rf_unprotect(NP);
+          Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
+        }
+        // Loop starting from the end of the previous run-length
+        for (int j = run_start; j != run_end; ++j){
+          // Check that order value is valid
+          if (has_order){
+            oi = p_o[j] - 1;
+            check_order_value(oi, o_rng, NP);
+          } else {
+            oi = j;
+          }
+          k = recycle_lag ? p_lag[oi % lag_size] : lag1;
+          if (k >= 0){
+            p_out[oi] = (j - run_start) >= k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
+          } else {
+            p_out[oi] = (j - run_end) < k ? p_x[has_order ? p_o[j - k] - 1 : j - k] : fill_value;
+          }
         }
       }
+      if (run_end != size){
+        Rf_unprotect(NP);
+        Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
+      }
     }
-    if (run_end != size){
-      Rf_unprotect(n_protections);
-      Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
-    }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
     break;
   }
   case STRSXP: {
@@ -627,9 +599,9 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     }
     const SEXP *p_x = STRING_PTR_RO(x);
     SEXP fill_value = Rf_protect(fill_size >= 1 ? Rf_asChar(fill) : NA_STRING);
-    ++n_protections;
+    ++NP;
     out = Rf_protect(Rf_duplicate(x));
-    ++n_protections;
+    ++NP;
     for (int i = 0; i != rl_size; ++i){
       run_start = run_end; // Start at the end of the previous run
       rl = has_rl ? p_rl[i] : size; // Current run-length
@@ -637,13 +609,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
 
       // If any run-lengths are negative (or NA) we stop
       if (rl < 0){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("All run lengths must be non-NA and >= 0");
       }
 
       // If the cumulative run-length exceeds length(x) we stop
       if (run_end > size){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
       // Loop starting from the end of the previous run-length
@@ -651,7 +623,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
         // Check that order value is valid
         if (has_order){
           oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
+          check_order_value(oi, o_rng, NP);
         } else {
           oi = j;
         }
@@ -664,10 +636,10 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       }
     }
     if (run_end != size){
-      Rf_unprotect(n_protections);
+      Rf_unprotect(NP);
       Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
     }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
     break;
   }
   case CPLXSXP: {
@@ -677,13 +649,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     }
     Rcomplex *p_x = COMPLEX(x);
     SEXP fill_sexp = Rf_protect(Rf_allocVector(CPLXSXP, 1));
-    ++n_protections;
+    ++NP;
     Rcomplex *p_fill = COMPLEX(fill_sexp);
     p_fill[0].i = NA_REAL;
     p_fill[0].r = NA_REAL;
     Rcomplex fill_value = fill_size >= 1 ? Rf_asComplex(fill) : COMPLEX(fill_sexp)[0];
     out = Rf_protect(Rf_duplicate(x));
-    ++n_protections;
+    ++NP;
     for (int i = 0; i != rl_size; ++i){
       run_start = run_end; // Start at the end of the previous run
       rl = has_rl ? p_rl[i] : size; // Current run-length
@@ -691,13 +663,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
 
       // If any run-lengths are negative (or NA) we stop
       if (rl < 0){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("All run lengths must be non-NA and >= 0");
       }
 
       // If the cumulative run-length exceeds length(x) we stop
       if (run_end > size){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
       // Loop starting from the end of the previous run-length
@@ -705,7 +677,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
         // Check that order value is valid
         if (has_order){
           oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
+          check_order_value(oi, o_rng, NP);
         } else {
           oi = j;
         }
@@ -718,10 +690,10 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       }
     }
     if (run_end != size){
-      Rf_unprotect(n_protections);
+      Rf_unprotect(NP);
       Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
     }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
     break;
   }
   case RAWSXP: {
@@ -732,9 +704,9 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     Rbyte *p_x = RAW(x);
     SEXP raw_sexp = Rf_protect(Rf_coerceVector(fill, RAWSXP));
     Rbyte fill_value = fill_size == 0 ? RAW(Rf_ScalarRaw(0))[0] : RAW(raw_sexp)[0];
-    ++n_protections;
+    ++NP;
     out = Rf_protect(Rf_duplicate(x));
-    ++n_protections;
+    ++NP;
     for (int i = 0; i != rl_size; ++i){
       run_start = run_end; // Start at the end of the previous run
       rl = has_rl ? p_rl[i] : size; // Current run-length
@@ -742,13 +714,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
 
       // If any run-lengths are negative (or NA) we stop
       if (rl < 0){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("All run lengths must be non-NA and >= 0");
       }
 
       // If the cumulative run-length exceeds length(x) we stop
       if (run_end > size){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
       // Loop starting from the end of the previous run-length
@@ -756,7 +728,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
         // Check that order value is valid
         if (has_order){
           oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
+          check_order_value(oi, o_rng, NP);
         } else {
           oi = j;
         }
@@ -769,10 +741,10 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       }
     }
     if (run_end != size){
-      Rf_unprotect(n_protections);
+      Rf_unprotect(NP);
       Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
     }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
     break;
   }
   case VECSXP: {
@@ -780,7 +752,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     int size = Rf_length(x);
     const SEXP *p_x = VECTOR_PTR_RO(x);
     out = Rf_protect(Rf_allocVector(VECSXP, size));
-    ++n_protections;
+    ++NP;
     SHALLOW_DUPLICATE_ATTRIB(out, x);
     for (int i = 0; i < size; ++i){
       SET_VECTOR_ELT(out, i, cpp_lag2(p_x[i], lag, order, run_lengths, fill, true));
@@ -792,9 +764,9 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
     }
     const SEXP *p_x = VECTOR_PTR_RO(x);
     SEXP fill_value = Rf_protect(VECTOR_ELT(Rf_coerceVector(fill_size >= 1 ? fill : R_NilValue, VECSXP), 0));
-    ++n_protections;
+    ++NP;
     out = Rf_protect(Rf_allocVector(VECSXP, size));
-    ++n_protections;
+    ++NP;
     for (int i = 0; i != rl_size; ++i){
       run_start = run_end; // Start at the end of the previous run
       rl = has_rl ? p_rl[i] : size; // Current run-length
@@ -802,13 +774,13 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
 
       // If any run-lengths are negative (or NA) we stop
       if (rl < 0){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("All run lengths must be non-NA and >= 0");
       }
 
       // If the cumulative run-length exceeds length(x) we stop
       if (run_end > size){
-        Rf_unprotect(n_protections);
+        Rf_unprotect(NP);
         Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
       }
       // Loop starting from the end of the previous run-length
@@ -816,7 +788,7 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
         // Check that order value is valid
         if (has_order){
           oi = p_o[j] - 1;
-          check_order_value(oi, o_rng, n_protections);
+          check_order_value(oi, o_rng, NP);
         } else {
           oi = j;
         }
@@ -829,18 +801,18 @@ SEXP cpp_lag2(SEXP x, SEXP lag, SEXP order, SEXP run_lengths, SEXP fill, bool re
       }
     }
     if (run_end != size){
-      Rf_unprotect(n_protections);
+      Rf_unprotect(NP);
       Rf_error("sum(run_lengths) must be equal to length(x) (%d)", size);
     }
-    LAG_R_NAMES;
+    CHEAPR_LAG_R_NAMES;
   }
   break;
   }
   default: {
-    Rf_unprotect(n_protections);
+    Rf_unprotect(NP);
     Rf_error("%s cannot handle an object of type %s", __func__, Rf_type2char(TYPEOF(x)));
   }
   }
-  Rf_unprotect(n_protections);
+  Rf_unprotect(NP);
   return out;
 }
